"""
This module contains a few small sphinx extensions.
They are mainly used to help with the generation
of BPS's own documentation, but some other projects
use them as well, so they are kept here.
"""
#=============================================================================
# imports
#=============================================================================
# core
import re
import os.path
import sys
# site
# WARNING: since this is imported by setup.py, it's critical the toplevel
#          not depend on any imports outside of stdlib & setuptools
# local
__all__ = [
    # constants
    "__version__",
    "std_exts",
    "all_exts",

    # public utility functions
    "get_theme_dir",
    "get_version",

    # internal helpers
    "is_cloud_theme",
]

#=============================================================================
# constants
#=============================================================================
__version__ = '1.9.4'

# names of standard cloud extensions used by most cloud themes
std_exts = [
    'cloud_sptheme.ext.autodoc_sections',
    'cloud_sptheme.ext.autoattribute_search_bases',
    'cloud_sptheme.ext.docfield_markup',
    'cloud_sptheme.ext.escaped_samp_literals',
    'cloud_sptheme.ext.index_styling',
    'cloud_sptheme.ext.relbar_links',
    'cloud_sptheme.ext.table_styling',
]

# names of all cloud extensions
all_exts = std_exts + [
    'cloud_sptheme.ext.issue_tracker',
]

#=============================================================================
# public helpers
#=============================================================================
_root_dir = os.path.abspath(os.path.dirname(__file__))

def get_theme_dir():
    """Returns path to directory containing this package's Sphinx themes.

    .. deprecated:: 1.7

        As of Sphinx 1.2, this is passed to Sphinx via a ``setup.py`` entry point,
        and no longer needs to be included in your documentation's ``conf.py``.
    """
    return os.path.join(_root_dir, "themes")

def get_version(release):
    """Derive short version string from longer 'release' string.

    This is quick helper which takes a project's ``release`` string,
    and generates the shortened ``version`` string required by ``conf.py``.
    Usage example for ``conf.py``::

        import cloud_sptheme as csp

        ...

        # The version info for the project you're documenting
        from myapp import __version__ as release
        version = csp.get_version(release)
    """
    return re.match("(\d+\.\d+)", release).group(1)

#=============================================================================
# misc internal helpers
#=============================================================================
def is_cloud_theme(name):
    """[hack] internal helper to check if theme accepts cloud theme options"""
    return os.path.isfile(os.path.join(get_theme_dir(), name, "theme.conf"))

#=============================================================================
# eof
#=============================================================================



"""
cloud_sptheme.ext.relbar_toc -- adds TOC entry (and other links) to relbar.
NOTE: this has been renamed to 'relbar_links'. this module is left as a legacy alias.
"""
#=========================================================
# imports
#=========================================================
# core
from __future__ import absolute_import, division, print_function
import warnings
# pkg
from . import relbar_links
# local
__all__ = [
    "setup"
]

#=========================================================
# sphinx entrypoint
#=========================================================
def setup(app):
    """
    setup extension for sphinx run
    """
    # TODO: decide when to remove this, and promote to "DeprecationWarning"
    warnings.warn("'cloud_sptheme.ext.relbar_toc' is deprecated, "
                  "use 'cloud_sptheme.ext.relbar_links' instead",
                  PendingDeprecationWarning)
    return relbar_links.setup(app)

#=========================================================
# eof
#=========================================================



"""
cloud_sptheme.utils -- internal helper utilities
"""
#=============================================================================
# imports
#=============================================================================
# core
from functools import update_wrapper
import logging; log = logging.getLogger(__name__)
import os
import sys
# site
from sphinx.util.fileutil import copyfile
# pkg
# local
__all__ = [
    # py2/3 compat
    'PY2', 'PY3', 'u', 'ru',
    'unicode', 'unicode_or_str',

    # monkeypatch helpers
    "patchapplier",
    "monkeypatch",
]

#=============================================================================
# internal py2/3 compat helpers
#=============================================================================
PY2 = sys.version_info < (3,0)
PY3 = not PY2

# FIXME: these aren't very rigorous / correct, but they work for current purposes.
if PY2:
    def u(s):
        return s.decode("unicode_escape")
    def ru(s):
        return s.decode("ascii")
    unicode = unicode
    unicode_or_str = (unicode, str)
else:
    def u(s):
        return s
    ru = u
    unicode = str
    unicode_or_str = (str,)

#=============================================================================
# monkeypatch helpers
#=============================================================================
def patchapplier(func):
    """
    function decorator to help functions that apply a monkeypatch.
    makes them only run once.
    """
    def wrapper():
        if wrapper.patched:
            return False
        func()
        wrapper.patched = True
        logging.getLogger(func.__module__).debug("%s: patch applied", func.__name__)
        return True
    wrapper.patched = False
    update_wrapper(wrapper, func)
    return wrapper

def monkeypatch(target, name=None):
    """
    helper to monkeypatch another object.
    the decorated function is wrapped around the existing function in
    :samp:`target.{name}`, and used to replace it.

    **name** defaults to the name of the function being decorated.

    the original value is passed in as the first positional argument to the function.
    """
    def builder(func):
        attr = name or func.__name__
        wrapped = getattr(target, attr)
        def wrapper(*args, **kwds):
            return func(wrapped, *args, **kwds)
        update_wrapper(wrapper, wrapped)
        wrapper.__wrapped__ = wrapped # not set by older update_wrapper() versions
        setattr(target, attr, wrapper)
        return func # return orig func so we can use it again
    return builder

#=============================================================================
# sphinx helpers
#=============================================================================

def _patch_html_extra_static(builder):
    """
    monkeypatch hook for add_static_file() to use
    """
    config = builder.config
    if hasattr(config, "html_extra_static"):
        return

    @monkeypatch(builder)
    def copy_static_files(wrapped):
        wrapped()
        # NOTE: code modeled after copy_static_files()...
        ctx = builder.globalcontext.copy()
        for source, target in config.html_extra_static:
            # 'source' existence should already be checked by add_static_file(),
            # 'target' should already be abspath w/in outdir
            copyfile(source, target)

    config.html_extra_static = []


def add_static_file(builder, source, name=None, stylesheet=False,
                    javascript=False):
    """
    monkeypatch sphinx's html builder to include specified static file.
    
    will copy file to :samp:`{outdir}/_static/{name}`.

    :param builder:
        ref to HTMLBuilder
        
    :param source: 
        path to source file
    
    :param name:
        name for target file; defaults to base name of source file.
        
    :param stylesheet:
        register file as additional stylesheet
    """
    _patch_html_extra_static(builder)

    if not os.path.exists(source):
        builder.warn('static asset %r does not exist' % source)
        return

    if not name:
        name = os.path.basename(source)

    target = os.path.join(builder.outdir, "_static", name)

    if stylesheet:
        builder.app.add_stylesheet(name)

    if javascript:
        builder.app.add_javascript(name)

    builder.config.html_extra_static.append((source, target))

#=============================================================================
# eof
#=============================================================================



# -*- coding: utf-8 -*-
"""
Sphinx configuration file for the cloud_sptheme documentation.

This file is execfile()d with the current directory set to its containing dir.
Note that not all possible configuration values are present in this
autogenerated file. All configuration values have a default; values that are
commented out serve to show the default.

This honors the following sphinx tags (passed via -t or $SPHINX_BUILD_TAGS):
* for-pypi -- generate special version to upload to pypi
"""
#=============================================================================
# environment setup
#=============================================================================
import sys, os

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
##sys.path.insert(0, os.path.abspath('.'))

# make sure root of source dir in sys.path
sys.path.insert(0, os.path.abspath(os.pardir))

# read env var for tags, needed since "python setup.py build_sphinx"
# doesn't support sphinx-build's "-t" option.
for _tag in os.environ.get("SPHINX_BUILD_TAGS", "").split():
    tags.add(_tag)

#=============================================================================
# imports
#=============================================================================
# import cloud_sptheme for themes, etc
import cloud_sptheme as csp

#=============================================================================
# General configuration
#=============================================================================

# If your documentation needs a minimal Sphinx version, state it here.
needs_sphinx = '1.4'

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    # standard sphinx extensions
    'sphinx.ext.autodoc',
    'sphinx.ext.todo',

    # 3rd party extensions
    'sphinxcontrib.fulltoc',

    # cloud's extensions
    'cloud_sptheme.ext.autodoc_sections',
    'cloud_sptheme.ext.relbar_links',
    'cloud_sptheme.ext.escaped_samp_literals',
    'cloud_sptheme.ext.issue_tracker',
    'cloud_sptheme.ext.table_styling',
    'cloud_sptheme.ext.role_index',  # NOTE: used only to provide example role index
    ]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The encoding of source files.
source_encoding = 'utf-8'

# The master toctree document.
master_doc = 'contents'

# The frontpage document.
index_doc = 'index'

# General information about the project.
project = "Cloud Sphinx Theme"
author = "Assurance Technologies, LLC"
copyright = "2010-2012, " + author

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.

# release: The full version, including alpha/beta/rc tags.
# version: The short X.Y version.
from cloud_sptheme import __version__ as release
version = csp.get_version(release)

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
##language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
##today = ''
# Else, today_fmt is used as the format for a strftime call.
##today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = [
    # may revise/remove this
    "lib/cloud_sptheme.make_helper.rst",
]

# The reST default role (used for this markup: `text`) to use for all documents.
##default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
##add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
##show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
modindex_common_prefix = [ "cloud_sptheme." ]

if tags.has("for-pypi"):

    extensions.append('cloud_sptheme.ext.auto_redirect')
    # auto_redirect_subject = "The official cloud_sptheme documentation"
    auto_redirect_domain_url = "https://cloud-sptheme.readthedocs.io"
    auto_redirect_domain_root = "/en/latest"

#=============================================================================
# Options for all output
#=============================================================================
todo_include_todos = True
keep_warnings = True
issue_tracker_url = "bb:ecollins/cloud_sptheme"

#=============================================================================
# Options for HTML output
#=============================================================================

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = os.environ.get("SPHINX_THEME") or 'cloud'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {}
if csp.is_cloud_theme(html_theme):
    html_theme_options.update(
        roottarget=index_doc,
        borderless_decor=True,
        sidebarwidth="3in",
        hyphenation_language="en",
    )

# Add any paths that contain custom themes here, relative to this directory.
html_theme_path = [csp.get_theme_dir()]

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
html_title = "%s v%s Documentation" % (project, release)

# A shorter title for the navigation bar.  Default is the same as html_title.
html_short_title = "%s %s Documentation" % (project, version)

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = os.path.join("_static", "masthead.png")

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
##html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
##html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
html_sidebars = {'**': ['searchbox.html', 'globaltoc.html']}

# Additional templates that should be rendered to pages, maps page names to
# template names.
##html_additional_pages = {}

# If false, no module index is generated.
##html_domain_indices = True

# If false, no index is generated.
##html_use_index = True

# If true, the index is split into individual pages for each letter.
##html_split_index = False

# If true, links to the reST sources are added to the pages.
##html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
##html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
##html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
##html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
##html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = project + 'Doc'

#=============================================================================
# Options for LaTeX output
#=============================================================================

# The paper size ('letter' or 'a4').
##latex_paper_size = 'letter'

# The font size ('10pt', '11pt' or '12pt').
##latex_font_size = '10pt'

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [
  (index_doc, project + '.tex', project + ' Documentation',
   author, 'manual'),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
##latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
##latex_use_parts = False

# If true, show page references after internal links.
##latex_show_pagerefs = False

# If true, show URL addresses after external links.
##latex_show_urls = False

# Additional stuff for the LaTeX preamble.
##latex_preamble = ''

# Documents to append as an appendix to all manuals.
##latex_appendices = []

# If false, no module index is generated.
##latex_domain_indices = True

#=============================================================================
# Options for manual page output
#=============================================================================

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (index_doc, project, project + ' Documentation',
     [author], 1)
]

#=============================================================================
# EOF
#=============================================================================



"""cloud_sptheme.ext.table_styling -- add directives for styling tables"""
#=============================================================================
# imports
#=============================================================================
# core
from cloud_sptheme.utils import PY3
if PY3:
    from itertools import zip_longest as izip_longest
else:
    # FIXME: not present in py25
    from itertools import izip_longest
import os
# site
from docutils import nodes
from docutils.parsers.rst import directives
from docutils.parsers.rst.directives.tables import RSTTable
from sphinx.builders.html import StandaloneHTMLBuilder
try:
    # sphinx 1.5 has it's own RSTTable, inherit from that if available.
    from sphinx.directives.patches import RSTTable
except ImportError:
    pass
# pkg
from cloud_sptheme import __version__
from cloud_sptheme import _root_dir, is_cloud_theme
from cloud_sptheme.utils import add_static_file
# local
__all__ = [
    "setup",
]

#=============================================================================
# constants
#=============================================================================

# name of key controlling whether css file is included
EMBED_KEY = "table_styling_embed_css"

# name of key controlling css class name
CLASS_KEY = "table_styling_class"

#=============================================================================
# field option parsers
#=============================================================================
def _split_argument_list(argument):
    if "," in argument:
        return argument.split(",")
    else:
        return argument.split()

def _parse_argument_map(argument, argmap, param):
    args = _split_argument_list(argument)
    if len(args) == 1 and all(c in argmap for c in args[0]):
        args = args[0]
    def norm(arg):
        try:
            return argmap[arg]
        except KeyError:
            raise ValueError("invalid %s: %r" % (param, arg))
    return [norm(arg) for arg in args]

_alignment_map = dict(
    l="left",
    r="right",
    c="center",
    j="justify",
    left="left",
    right="right",
    center="center",
    centered="center", # compat alias
    justify="justify",
    justified="justify", # compat alias
)

def alignment_list(argument):
    """convert into list of alignment options.
    raise ``ValueError`` if no args found, or invalid strings.
    """
    return _parse_argument_map(argument, _alignment_map, "alignment")

_bool_map = {"true": True, "t": True,
             "yes": True, "y": True,
             "false": False, "f": False,
             "no": False, "n": False,
             }

def bool_list(argument):
    """convert to list of true/false values"""
    return _parse_argument_map(argument, _bool_map, "boolean value")

def class_option_list(argument):
    """convert to list of list of classes"""
    args = _split_argument_list(argument)
    return [directives.class_option(arg) for arg in args]

_divider_map = {
    "0": "no",
    "1": "single",
    "2": "double",
    "none": "no",
    "single": "single",
    "double": "double",
    }

def divider_list(argument):
    return _parse_argument_map(argument, _divider_map, "divider style")

#=============================================================================
# replacement for table directive
#=============================================================================

class ExtendedRSTTable(RSTTable):
    # TODO: could have this auto-generate tabularcolumns directive for latex,
    #       based on alignment, dividers, and wrapping settings:
    #           |,|| for dividers
    #           l,c,r - align, nowrap
    #           L,C,R,J - align, wrap
    #       XXX: how to handle widths for tabularcolumns?

    option_spec = RSTTable.option_spec.copy()
    option_spec.update({
        # class, name already present
        'header-columns': directives.nonnegative_int,
        # TODO: column-widths: support limited set of units (em/in/%)
        #       expressable under both css & latex
        'widths': directives.positive_int_list,
        'column-alignment': alignment_list,
        'header-alignment': alignment_list,
        'column-wrapping': bool_list,
        'column-classes': class_option_list,
        'column-dividers': divider_list,
    })

    def run(self):
        result = RSTTable.run(self)
        if result and isinstance(result[0], nodes.table):
            self._update_table_classes(result[0])
        return result

    def _update_table_classes(self, table):
        # add global "table_styling_class"
        assert isinstance(table, nodes.table)
        config = self.state.document.settings.env.config
        classes = getattr(config, CLASS_KEY)
        if classes:
            for cls in classes.split():
                table['classes'].append(cls)

        # figure out how many header rows & columns
        options = self.options
        header_cols = options.get("header-columns") or 0
        widths = options.get("widths")

        # Parse column options into list of ColumnOptions records.
        # If any option is short, it will be padded with ``None`` values
        # to match the longest list.
        EMPTY = ()
        opts = tuple(izip_longest(
            options.get("column-alignment", EMPTY),
            options.get("header-alignment", EMPTY),
            options.get("column-wrapping", EMPTY),
            options.get("column-classes", EMPTY),
            options.get("column-dividers", EMPTY),
        ))
        NULL_OPTIONS = [None] * 5

        # try to find "tgroup" node holding thead/tbody etc
        def locate(cls):
            for child in table.children:
                if isinstance(child, cls):
                    return child
            return None
        tgroup = locate(nodes.tgroup)
        if not tgroup:
            return

        # loop through content of 'tgroup' node --
        # should be 0+ 'colspec' nodes, 0-1 thead node, and 0-1 tbody nodes.
        # NOTE: using 'cgroup_idx' as hack to track colgroup index
        cgroup_idx = 0
        for child in tgroup:
            # handle colspec entry (corresponding to COLGROUP)
            # set group width & mark "header" columns
            if isinstance(child, nodes.colspec):
                if widths and cgroup_idx < len(widths):
                    child['colwidth'] = widths[cgroup_idx]
                if cgroup_idx < header_cols:
                    child['stub'] = 1
                cgroup_idx += 1
                continue

            # otherwise have thead / tbody --
            # loop through each row in child, adding classes as appropriate
            assert isinstance(child, (nodes.thead, nodes.tbody))
            for row in child.children:

                # check if we're in a header
                is_header_row = isinstance(child, nodes.thead)

                # Add alignment and wrap classes to each column (would add to
                # colspec instead, but html doesn't inherit much from colgroup)
                # We iterate trough the list of table cells in the row while
                # maintaining the column format index in parallel
                assert isinstance(row, nodes.row)
                colidx = 0  # visible idx, to account for colspans
                prev_col_classes = None
                for idx, col in enumerate(row):

                    # get align/wrap options for column
                    try:
                        col_opts = opts[colidx]
                    except IndexError:
                        col_opts = NULL_OPTIONS
                    align, header_align, wrap, clist, divider = col_opts

                    # let header values override defaults
                    if is_header_row:
                        align = header_align
                    if align is None:
                        align = "left"

                    # add alignment / wrapping classes for column
                    assert isinstance(col, nodes.entry)
                    classes = col['classes']
                    if align:
                        classes.append(align + "-align")
                    if wrap is False:
                        classes.append("nowrap")
                    if clist:
                        classes.extend(clist)

                    if divider:
                        # add divider class for left side of this column,
                        classes.append(divider + "-left-divider")

                        # same divider type to right side of previous column,
                        # so that css can bind to either.
                        if prev_col_classes is not None:
                            prev_col_classes.append(divider + "-right-divider")

                    # update prev_col_classes for next pass
                    prev_col_classes = classes

                    # Now it's the time to advance the column format index.
                    # If we have a cell spans across the columns, we add the
                    # value of 'morecols', so we get the proper right divider
                    # for the current cell and the proper format index for the
                    # next loop iteration
                    colidx += 1 + col.get("morecols", 0)

                # add divider to right side of last column
                if prev_col_classes and colidx < len(opts):
                    align, header_align, wrap, clist, divider = opts[colidx]
                    if divider:
                        prev_col_classes.append(divider + "-right-divider")

#=============================================================================
# patch builder to copy css file (if needed)
#=============================================================================

css_path = os.path.join(_root_dir, "ext", "static", "table_styling.css")

def prepare_builder(app):
    # make sure needed css styling gets included when building html
    builder = app.builder
    if not isinstance(builder, StandaloneHTMLBuilder):
        return
    value = getattr(app.config, EMBED_KEY)
    if value is None:
        value = not is_cloud_theme(app.config.html_theme)
    if not value:
        return

    # add custom css stylesheet
    add_static_file(builder, css_path, stylesheet=True)

#=============================================================================
# register extension
#=============================================================================
def setup(app):
    # whether it will embed raw css
    app.add_config_value(EMBED_KEY, None, "html")

    # default class to add to all styled tables
    app.add_config_value(CLASS_KEY, "styled-table", "html")

    # HACK: since sphinx 1.5, issues warning about replacing RSTTable.
    #       we know that's happening, so pop it out of registry first.
    #       (but ONLY what's there wasn't already patched)
    directive_map = directives._directives
    if directive_map.get("table") is RSTTable:
        directive_map.pop("table")

    # replace existing table directive with custom one
    app.add_directive("table", ExtendedRSTTable)

    # add extra resources
    app.connect("builder-inited", prepare_builder)

    # identifies the version of our extension
    return {'version': __version__}

#=============================================================================
# eof
#=============================================================================



"""
======================================================
:mod:`cloud_sptheme.ext.role_index` -- JSON Role Index
======================================================

.. versionadded:: 1.8

Overview
========
This sphinx extension for the HTML builder places a file called ``roleindex.json``
in the output directory along with the documentation.  This file contains
a dictionary which mapping all reference ids (made via ``:ref:`` or ``:doc:``) to
a 3-element list of ``(page, anchor, title)``

This is useful for integrating a Sphinx manual into the online help of web application --
context-specific help can be provided by looking up a predefined ``:ref:`` tag within
the json file, and redirecting the user to the appropriate page and anchor location.

As an example, see the `roleindex.json <../roleindex.json>`_ for this documentation.

.. warning::

    This extension should be considered "beta" quality.
    It was recently written, may have unknown issues, and may need to be revised.

Configuration
=============
This extension reads the following ``conf.py`` options:

    ``role_index_style``

        style of json output -- can be ``"compact"`` (the default), or ``"pretty"`` (for debugging).
"""
#=========================================================
# imports
#=========================================================
# core
from __future__ import absolute_import, division, print_function
import codecs
import logging
log = logging.getLogger(__name__)
import json
import os
# site
from sphinx.locale import _TranslationProxy
# pkg
from cloud_sptheme import __version__
# local
__all__ = [
    "setup",
]
#=========================================================
# hooks
#=========================================================
def write_role_index(app):
    """
    helper which generates ``"roleindex.json"`` for document.
    """
    if app.buildername != "html":
        return

    style = app.config.role_index_style or "compact"

    index = {}

    # add 'ref' role mappings
    index['ref'] = dict(
        (key, (page, anchor, label))
        for key, (page, anchor, label)
        in app.env.domaindata['std']['labels'].items()
    )

    # add 'doc' role mappings
    index['doc'] = dict(
        (doc.lower(), (doc, "", title.astext()))
        for doc, title in app.env.titles.items()
    )

    # create composite 'all' mapping
    index['any'] = index['doc'].copy()
    index['any'].update(index['ref'])

    # write to file
    def encode(value):
        if isinstance(value, _TranslationProxy):
            return str(value)
        raise TypeError("can't serialize value: %r (type=%r)" % (value, type(value)))

    target = os.path.join(app.builder.outdir, "roleindex.json")
    kwds = dict(sort_keys=True, separators=(",", ":"), default=encode)
    if style == "compact":
        pass
    elif style == "pretty":
        kwds.update(separators=(", ", ": "), indent=True)
    else:
        raise ValueError("unknown role_index_style: %r" % style)
    with codecs.open(target, "w", encoding="utf-8") as fh:
        json.dump(index, fh, **kwds)

#=========================================================
# sphinx entrypoint
#=========================================================
def setup(app):

    app.add_config_value('role_index_style', None, 'html')

    # XXX: this probably isn't right event to attach to...
    #      but want to run during writing phase, after all pages have been scanned into tree,
    #      and this seems like close to the right place.
    def wrapper(app):
        write_role_index(app)
        return []
    app.connect('html-collect-pages', wrapper)

    # identifies the version of our extension
    return {'version': __version__}

#=========================================================
# eof
#=========================================================



"""cloud_sptheme.ext.autodoc_sections - support ReST sections in docstrings"""
#=============================================================================
# imports
#=============================================================================
# core
import inspect
import re
import logging; log = logging.getLogger(__name__)
# site
# pkg
from cloud_sptheme import __version__
from cloud_sptheme.utils import u, patchapplier, monkeypatch
# local
__all__ = [
    "setup",
]

#=============================================================================
# internal helpers used for hacking up sphinx. almost ashamed to be doing this.
#=============================================================================
def get_caller_value(module, var, rtype=None, code=None):
    """
    helper which looks for nearest ancestor in call stack
    which occurred in module, and return specified local variable.

    :param module:
        name of module to look for on stack

    :param var:
        name of local variable to return

    :param rtype:
        do optional type-check to ensure **name** has expected type.

    :param code:
        optionally match based on function name (``frame.f_code.co_name``)
        as well as module.
    """
    frame = None
    try:
        frame = inspect.currentframe().f_back
        while True:
            if ((frame.f_globals.get("__name__") == module) and
                (not code or frame.f_code.co_name == code)):
                    break
            frame = frame.f_back
            if not frame:
                raise RuntimeError("couldn't find module=%r, code=%r in call stack" %
                                   (module, code or '<any>'))
        value = frame.f_locals[var]
        if rtype and not isinstance(value, rtype):
            raise TypeError("%s: expected a %r instance: %r" % (var, rtype, value))
        return value
    finally:
        frame = None

#=============================================================================
# autodoc monkeypatches / hacks
#=============================================================================
@patchapplier
def _patch_sphinx():
    """
    helper which monkeypatches sphinx to install some of our hooks.
    """

    #----------------------------------------------------------------------
    # patch document.note_implicit_target() to look for _modify_new_desc_section
    # attribute, as signal that it should munge up node that's passed to it,
    # to represent new description-level section, rather than document-level section.
    # this flag is then set by RSTState.new_subsection() patch, below.
    # NOTE: ideally, all this action could be done by a hook w/in
    #       RSTState.new_subsection(), but would have to modify source.
    #----------------------------------------------------------------------
    from docutils.nodes import document, make_id

    @monkeypatch(document)
    def note_implicit_target(_wrapped, self, target, *args, **kwds):
        # use default behavior unless signal flag is set
        entry = self._modify_new_desc_section
        if not entry:
            return _wrapped(self, target, *args, **kwds)
        self._modify_new_desc_section = None

        # NOTE: target should be section() node we're modifying,
        #       as we just got called from RSTState.new_subsection(),
        #       which is what sets modify_new_desc_section flag.
        #       'entry' should be last item in memo.desc_stack
        #       (see below).

        # add our custom css classes for styling
        # NOTE: adding 'section-header' class to H<N> node,
        #       so that our css rules don't have to be duplicated for every H<N> value.
        target['classes'].append("desc-section")
        target['classes'].append("desc-section-%d" % entry['level'])
        target[0]['classes'].append("section-header")

        # for duration of call, modify settings.id_prefix to include
        # decription prefix in any auto-generated ids. this helpers
        # section names remaining unique even if used between classes
        # (e.g. common names such as 'Constructor Options')
        settings = self.settings
        orig = settings.id_prefix
        try:
            if entry['prefix']:
                if orig:
                    settings.id_prefix = u("%s-%s") % (settings.id_prefix, entry['prefix'])
                else:
                    settings.id_prefix = entry['prefix']
            return _wrapped(self, target, *args, **kwds)
        finally:
            settings.id_prefix = orig

    document._modify_new_desc_section = None

    #----------------------------------------------------------------------
    # patch RSTState.new_subsection() to generate sections nested within
    # a description. It reads ``memo.desc_stack`` to determine if it's within
    # a description. If set, this attr should be a list of dicts,
    # each entry representing a nested description (e.g. an ObjectDescription)
    # whose content is being parsed, most recent should be last.
    # Each entry should be a dict containing:
    #
    # If there are no description entries active, the normal behavior is used.
    #
    # Each dict should contain the following keys:
    #   * prefix -- None, or string to use as prefix for section identifiers.
    #               helps keep links unique w/in document.
    #   * signode -- signature node used to generate prefix (for debugging)
    #   * owner -- arbitary object (ObjectDescription in our case)
    #              which added this entry to list. intended as sanity check
    #              when popping entries back off stack.
    #   * level -- section level w/in declaration (autoset by code below)
    #----------------------------------------------------------------------
    from docutils.parsers.rst.states import RSTState

    @monkeypatch(RSTState)
    def new_subsection(_wrapped, self, *args, **kwds):
        desc_stack = getattr(self.memo, "desc_stack", None)
        if desc_stack:
            # after new_subsection() creates section node,
            # it will invoke document.note_implicit_target().
            # setting this attr signals our monkeypatch of that method (above)
            # to make changes to that node based on desc_stack entry.
            # NOTE: ideally, the note_implicit_target() monkeypatch,
            #       as well as this code, would be placed inside RSTState.new_subsection(),
            #       but that would require modifying sphinx's source :(
            entry = desc_stack[-1]
            entry['level'] = self.memo.section_level+1 # set level w/in description
            self.document._modify_new_desc_section = entry # enable note hack

        # hand off to real method
        return _wrapped(self, *args, **kwds)

    #----------------------------------------------------------------------
    # monkeypatch ObjectDescription.run() so that:
    # 1. before calling state.nested_parse(), we push a desc context
    #    onto state_machine.desc_context_stack (see above).
    # 2. when it does call state.nested_parse() the first time,
    #    ``match_titles=True`` gets set.
    #    FIXME: using a really awkward way to accomplish this :|
    # 3. pop our context off desc_context_stack when done.
    #----------------------------------------------------------------------
    from sphinx.directives import ObjectDescription
    from sphinx.addnodes import desc as DescNodeType

    @monkeypatch(ObjectDescription)
    def run(_wrapped_run, self):
        # ObjectDescription.before_content() will be invoked right before
        # run calls ``self.state.nested_parse()``. We take advantage of that
        # by wrapping the instance's before_content() call so that the last
        # thing is does is set up ``self.state`` the way we want.

        # NOTE: have to patch per-instance, since subclasses that override this
        #       don't tend to invoke super()
        @monkeypatch(self)
        def before_content(_wrapped_before):
            # let real method do all the setup it wants.
            _wrapped_before()

            #----------------------------------------------------------------------
            # need to figure out prefix to prepend to our description sections,
            # so their IDs are unique. for now, using signature node
            # ObjectDescription.run() has finished generating right before before_content()
            # was called. Unfortunately, it's not available via self,
            # so we have to reach into call stack to grab it...
            # NOTE: ideally we would do this in ObjectDescription.run().
            #----------------------------------------------------------------------
            node = get_caller_value("sphinx.directives", "node",
                                    rtype=DescNodeType, code="run")
            # FIXME: would like a more bullet-proof way of deriving our id prefix...
            signode = node.children[0]
            if signode.get("ids"):
                base = signode['ids'][0]
            elif signode.get("names"):
                base = signode['names'][0]
            else:
                base = signode.astext()
            prefix = re.sub("[^a-zA-Z0-9_.]+", "-", base).strip("-") + "-"

            # now that we've got that info, add our description context entry
            # to the stack
            memo = self.state.memo
            if not hasattr(memo, "desc_stack"):
                memo.desc_stack = []
            memo.desc_stack.append(dict( # see new_subsection() above for dict format
                prefix=prefix,
                owner=self,
                signode=signode,
                level=0,
            ))

            #----------------------------------------------------------------------
            # hack up ``state.nested_parse()`` so that the next time it's called,
            # 'match_titles=True' is set. that call should happen as soon as this
            # function returns back to ObjectDescription.run()
            # NOTE: ideally, we would just set match_titles=True within ObjectDescription.run()
            #----------------------------------------------------------------------
            state = self.state
            if not hasattr(state, "_set_next_match_titles_flag"):
                # state is persistent object, only want to patch it once.
                @monkeypatch(state)
                def nested_parse(_wrapped_parse, *args, **kwds):
                    if state._set_next_match_titles_flag:
                        kwds['match_titles'] = True
                        state._set_next_match_titles_flag = False
                    return _wrapped_parse(*args, **kwds)

            # signal our hack to set match_titles
            state._set_next_match_titles_flag = True

        @monkeypatch(self)
        def after_content(_wrapped_after):
            # remove our description context entry
            # NOTE: ideally would do this in ObjectDescription.run()
            desc = self.state.memo.desc_stack.pop()
            assert desc['owner'] is self, "sanity check failed"

            # let real method do it's work
            return _wrapped_after()

        # now invoke the real run() method.
        # as soon as it calls before_content(), our hack above will patch self.state.
        # after before_content() returns, real run method will call self.state.nested_parse(),
        # and invoke our patched version instead.
        return _wrapped_run(self)

    #----------------------------------------------------------------------
    # make autodoc invoke parse_nested_section_with_titles() for ALL objects
    # if this isn't done, autodoc generates paragraphs instead of sections.
    # this causes all nested content to be omitted
    # FIXME: why is the lack of this causing a problem? should track it down.
    #----------------------------------------------------------------------
    from sphinx.ext.autodoc import Documenter
    Documenter.titles_allowed = True

    #----------------------------------------------------------------------
    # finally, monkeypatch DocFieldTransformer.transform_all()
    # so that it transforms doc fields  within one of our nested sections
    # (default code only looks at top-level nodes)
    #
    # FIXME: find a cleaner way to do this :|
    #----------------------------------------------------------------------
    from sphinx.util.docfields import DocFieldTransformer
    from docutils.nodes import section

    @monkeypatch(DocFieldTransformer)
    def transform_all(_wrapped, self, node):
        # transform immediate node contents like normal
        _wrapped(self, node)

        # our nested sections show up as definition lists,
        # so make sure transform_all is also invoked for the contents
        # of any definition list
        for child in node:
            if isinstance(child, section):
                _wrapped(self, child.children)

    #----------------------------------------------------------------------
    # sigh. done monkeypatching.
    #----------------------------------------------------------------------

#=============================================================================
# docstring mangling
#=============================================================================
def trim_module_header(app, what, name, obj, options, lines):
    """
    helper to remove one-line description from top of module (if preset).
    """
    if what != "module":
        return
    _title_re = re.compile(r"""
        ^ \s*
        ( {0} \s* -- \s* )?
        [a-z0-9 _."']*
        $
    """.format(re.escape(name)), re.X|re.I)
    if len(lines) > 1 and _title_re.match(lines[0]) and lines[1].strip() == '':
        del lines[:2]

#=============================================================================
# sphinx extension entrypoint
#=============================================================================
def setup(app):
    # don't patch sphinx unless this extension is actually in use
    _patch_sphinx()

    # clean up leading bit of module docstring
    app.connect('autodoc-process-docstring', trim_module_header)

    # identifies the version of our extension
    return {'version': __version__}

#=============================================================================
# documentation helper
#
# NOTE: this function doesn't actually do anything,
#       it exists to test this extension's behavior as part of docs/cloud_theme_test
#=============================================================================
def _doctestfunc():
    """
    The :mod:`~cloud_sptheme.ext.autodoc_sections` extension should generate
    nested sections as found within object docstrings.

    Nested Section
    ==============

    :param arg: xxx

    .. attribute:: foo

        bar

    These sections can in turn contain others:

    Child Section
    -------------

    Which allows breaking long class docstrings up in meaningful ways.

    Child Section 2
    ---------------

    And more content

    Nested Section 2
    ================

    end of class
    """
    pass

#=============================================================================
# eof
#=============================================================================



"""
==============================================================
:mod:`cloud_sptheme.ext.page_only` -- Exclude Pages from Build
==============================================================

.. versionadded:: 1.8

Overview
========
This extension provides a ``.. page-only::`` directive,
which allows conditionally omitting entire pages from the document based on build tags.

The first line of a page should be a ``.. page-only:: <expr>`` directive;
where ``<expr>`` is a tag, or simple tag expression, just like ``.. only::``.
If the expression evaluates to False, the page won't be included in the final build
(and should be silently omitted from any TOC lists).

E.g. ``.. page-only:: draft and html``
"""
#=============================================================================
# imports
#=============================================================================
# core
from __future__ import absolute_import, division, print_function
import io
import logging
log = logging.getLogger(__name__)
import re
from weakref import WeakKeyDictionary
# site
from docutils.parsers.rst import Directive
import sphinx.addnodes
# pkg
from cloud_sptheme import __version__
from cloud_sptheme.utils import patchapplier, monkeypatch
# local
__all__ = [
    "setup",
]

#=============================================================================
# patching
#=============================================================================
@patchapplier
def _patch_sphinx():

    #
    # to get files list right, best to modify found_files before env.get_outdated_files()
    # gets a hold of them, and starts looking for changes.
    #
    from sphinx.environment import BuildEnvironment
    @monkeypatch(BuildEnvironment)
    def get_outdated_files(__wrapped__, env, config_changed):
        # call out hook to modify found_docs to remove omitted entries,
        # and store them in env.omitted_by_page_only
        strip_omitted_docs(env)

        # hand off to orig function
        return __wrapped__(env, config_changed)

    #
    # patch TocTree.run() so it doesn't warning about broken toc entries.
    #
    from sphinx.directives.other import TocTree
    @monkeypatch(TocTree)
    def run(__wrapped__, self):
        env = self.state.document.settings.env
        omitted_docs = env.omitted_by_pageonly
        found_docs = env.found_docs
        try:
            # temporarily restore found_docs so TocTree doesn't throw warning.
            # NOTE: this causes broken toc entries to be generated (rather than warnings).
            #       these are cleaned up by remove_toc_entries() below
            env.found_docs = found_docs | omitted_docs
            return __wrapped__(self)
        finally:
            env.found_docs = found_docs

#=============================================================================
# hooks
#=============================================================================

#: regex used to detect page-only directive
_header_re = re.compile(r"^\.\.\s+page-only::\s+(?P<expr>.+)\s*$")


#: as hack, using this to store reference to app so we can resolve it w/in env monkeypatch
_env_to_app = WeakKeyDictionary()


def install_app_ref(app):
    """
    hack to get app ref w/in env.get_outdated_files() monkeypatch,
    called on 'builder-inited' event
    """
    _env_to_app[app.env] = app


def strip_omitted_docs(env):
    """
    helper invoked right before env.get_outdated_files() is called,
    scans for omitted files and modifies document set accordingly.

    called via monkeypatch before .get_outdated_files() is run,
    which happens during env.update(), shortly before 'env-get-outdated' event.
    """
    # scan through found docs, removing any that should be omitted
    app = _env_to_app[env]
    found_docs = env.found_docs
    omitted_docs = set()
    for doc in found_docs:
        # find first non-empty line
        with io.open(env.doc2path(doc), encoding="latin-1") as fh:
            line = ''
            for line in fh:
                if line.strip():
                    break

        # check if it's a PageOnly directive
        m = _header_re.match(line)
        if not m:
            continue

        # add page to omitted list if expression doesn't validate
        if not app.tags.eval_condition(m.group("expr")):
            omitted_docs.add(doc)

    if omitted_docs:
        app.info("%s pages omitted by page-only" % len(omitted_docs))

    # persist omitted list for TocTree.run() wrapper and remove_from_toc()
    env.omitted_by_pageonly = omitted_docs

    # remove omitted from 'found_docs', so get_outdated_files() doesn't see them
    found_docs.difference_update(omitted_docs)


class PageOnly(Directive):
    """
    Directive used to capture info.
    Notes that this doesn't actually do anything, job has already been done by find_omitted_docs().
    This is just here to prevent parse errors.
    """
    # NOTE: cloned from Author directive, should be ok...
    has_content = False
    required_arguments = 1
    optional_arguments = 0
    final_argument_whitespace = True
    option_spec = {}

    def run(self):
        # XXX: could issue warning if this is encountered anywhere but top of document.
        return []


def remove_from_toc(app, doctree):
    """
    remove omitted pages from TOC, let through by TocTree.run() monkeypatch
    called on doctree-read event.
    """
    omitted_docs = app.env.omitted_by_pageonly
    for toctreenode in doctree.traverse(sphinx.addnodes.toctree):
        entries = toctreenode['entries']
        # print("%r %r %r" % (doctree, toctreenode, entries))
        # dbgcon()
        idx = 0
        while idx < len(entries):
            entry = entries[idx]
            # entry is (label|None, doc) tuple...
            if str(entry[1]) in omitted_docs:
                del entries[idx]
            else:
                idx += 1

#=============================================================================
# sphinx entrypoint
#=============================================================================
def setup(app):
    _patch_sphinx()
    app.add_directive('page-only', PageOnly)
    app.connect('builder-inited', install_app_ref)
    app.connect('doctree-read', remove_from_toc)

    # identifies the version of our extension
    return {'version': __version__}

#=============================================================================
# eoc
#=============================================================================



"""helper for quick cross-platform makefile for sphinx

TODO: this was hacked up really quickly, could use lots of work.
"""
#===============================================================
#imports
#===============================================================
#core
from __future__ import print_function
import logging; log = logging.getLogger(__name__)
import os,sys
from string import Template
import subprocess
#pkg
#local
__all__ = [
    "SphinxMaker",
]

#===============================================================
#misc helpers
#===============================================================
def sub(fmt, **kwds):
    if not kwds:
            kwds = globals()
    return Template(fmt).substitute(**kwds)

#===============================================================
#fs helpers
#===============================================================
joinpath = os.path.join

def abspath(*args):
    return os.path.abspath(joinpath(*args))

if hasattr(os.path, "realpath"):
    def realpath(*args):
        return os.path.realpath(joinpath(*args))
else:
    #probably windows - fake it best we can
    def realpath(*args):
        return os.path.normcase(os.path.abspath(joinpath(*args)))

def pathdir(path):
    return os.path.split(path)[0]

def clearpath(path):
    "recursively remove all contents of dir, but leave dir"
    for root, dirs, files in os.walk(path, topdown=False):
        for name in files:
            os.remove(joinpath(root, name))
        for name in dirs:
            os.rmdir(joinpath(root, name))

def rmpath(path):
    "drecursively delete path"
    if os.path.exists(path):
        if os.path.isdir(path):
            clearpath(path)
            os.rmdir(path)
        else:
            os.remove(path)

def ensuredirs(path):
    "ensure specified directory & all parents exist"
    if not os.path.isdir(path):
        os.makedirs(path)

#===============================================================
#main class
#===============================================================
class SphinxMaker(object):
    #===============================================================
    #class attrs
    #===============================================================

    # You can subclass these variables
    SPHINXOPTS    = []
    SPHINXBUILD   = "sphinx-build"
    PAPER         = "letter"
    SERVEHTML_PORT = 8000

    # Paths
    BUILD = "_build"
    SOURCE = "."

    #internal opts
    PAPEROPT_a4     = ["-D","latex_paper_size=a4"]
    PAPEROPT_letter = ["-D","latex_paper_size=letter"]

    #: list of attrs to check os.environ for overriddes.
    env_vars = [ "SPHINXOPTS", "SPHINXBUILD", "PAPER", "SERVEHTML_PORT", "BUILD", "SOURCE" ]

    #===============================================================
    #instance attrs
    #===============================================================
    root_dir = None
    conf_file = None
    conf = None

    #===============================================================
    #frontend
    #===============================================================
    def __init__(self, root_dir=None, **kwds):
        #FIXME: this may not be properly flexible.
        if root_dir is None:
            root_dir = joinpath(sys.modules["__main__"].__file__, os.pardir)
        self.root_dir = abspath(root_dir)
        self.conf_file = joinpath(self.root_dir, "conf.py")
        if not os.path.exists(self.conf_file):
            raise RuntimeError("conf file not found in root: %r" %
                               (self.root_dir))

        #check environment for overrides, as well as constructor
        for key in self.env_vars:
            value = kwds.pop(key, None)
            value = os.environ.get(key, value)
            if value is not None:
                t = type(getattr(self,key))
                #FIXME: this is *real* hacked way to do type conversion
                if isinstance(t, str):
                    if isinstance(t, int): #for ints, eg SERVEHTML_PORT
                        value = int(t)
                    elif isinstance(t, list): #for list of arguments, eg SPHINXOPTS
                        #FIXME: should use proper quote escaping logic when we split :(
                        value = " ".split(value)
                setattr(self, key, value)

        # make all relative paths relative to root dir
        for name in ("BUILD", "SOURCE"):
            value = getattr(self, name)
            if not os.path.isabs(value):
                value = abspath(self.root_dir, value)
                setattr(self, name, value)

        if kwds:
            raise TypeError("unknown keywords: %r" % (kwds,))

    @classmethod
    def execute(cls, args=None, **kwds):
        return cls(**kwds).run(args)

    def run(self, args=None):
        if args is None:
            args = sys.argv[1:]
        os.chdir(self.root_dir) #due to relative paths like self.BUILD
        for arg in args:
            getattr(self,"target_"+arg)()

    #===============================================================
    #targets
    #===============================================================
    def target_help(self):
        print("Please use \`make <target>' where <target> is one of")
        print("  clean     remove all compiled files")
        print("  html      to make standalone HTML files")
        print("  servehtml to serve standalone HTML files on port 8000")
#        print("  pickle    to make pickle files")
#        print("  json      to make JSON files")
        print("  htmlhelp  to make HTML files and a HTML help project")
#        print("  latex     to make LaTeX files, you can set PAPER=a4 or PAPER=letter")
#        print("  changes   to make an overview over all changed/added/deprecated items")
#        print("  linkcheck to check all external links for integrity")

    def target_clean(self):
        rmpath(self.BUILD)

    def target_html(self):
        self.build("html")

    def target_htmlhelp(self):
        self.build("htmlhelp")

    def target_servehtml(self):
        path = realpath(self.BUILD, "html")
        port = self.SERVEHTML_PORT

        # try to use paste
        try:
            from paste.httpserver import serve
            from paste.urlparser import StaticURLParser
        except ImportError:
            # fall back to stdlib server
            if sys.version_info[0] >= 3:
                import http.server as s
                HTTPServer = s.HTTPServer
            else:
                import SimpleHTTPServer as s
                HTTPServer = s.BaseHTTPServer.HTTPServer

            os.chdir(path)
            print("Serving files from %r on port %r" % (path, port))

            HTTPServer(('',port), s.SimpleHTTPRequestHandler).serve_forever()
        else:
            serve(StaticURLParser(path), host="0.0.0.0", port=port)

    #TODO: support latex, pdf, etc...

    ##def target_latex(self):
    ##    build("latex")
    ##    print("Run \`make all-pdf' or \`make all-ps' in that directory to" \
    ##        "run these through (pdf)latex.")
    ##
    ##def target_pdf():
    ##    assert os.name == "posix", "pdf build support not automated for your os"
    ##    build("latex")
    ##    target = BUILD / "latex"
    ##    target.chdir()
    ##    subprocess.call(['make', 'all-pdf'])
    ##    print("pdf built")

    #===============================================================
    #helpers
    #===============================================================
    def build(self, name):
        BUILD = self.BUILD
        ALLSPHINXOPTS = self.get_sphinx_opts()

        dt = joinpath(BUILD, "doctrees")
        ensuredirs(dt)

        target = joinpath(BUILD, name)
        ensuredirs(target)

        rc = subprocess.call([self.SPHINXBUILD, "-b", name] + ALLSPHINXOPTS + [ target ])
        if rc:
            print("Sphinx-Build returned error, exiting.")
            sys.exit(rc)
        print("Build finished. The %s pages are in %r." % (name, target,))
        return target

    def get_paper_opts(self):
        return getattr(self,"PAPER_" + self.PAPER, [])

    def get_sphinx_opts(self):
        return ["-d", joinpath(self.BUILD, "doctrees")] + self.get_paper_opts() + self.SPHINXOPTS + [ self.SOURCE ]

    #===============================================================
    #eoc
    #===============================================================

class ProjectSphinxMaker(SphinxMaker):
    "SphinxMaker variant which more usefully integrates into setup.py of a python project"
    #TODO: make this read setup.cfg etc to see where build_sphinx *actually* puts things
    BUILD = os.path.join(os.pardir, "build", "sphinx")

#===============================================================
#eof
#===============================================================



"package containing various sphinx extensions"

#legacy aliases
from __future__ import absolute_import
from . import (autodoc_sections as nested_sections,
               index_styling as index_styles,
               )



"""
=================================================================
:mod:`cloud_sptheme.ext.relbar_links` - adds TOC entry to relbar
=================================================================

.. module:: cloud_sptheme.ext.relbar_toc
    :synopsis: adds TOC entry to right side of relbar

.. versionadded:: 1.8.1

Overview
========
This Sphinx extension allows you to insert links into
the navigation bar ("relbar") at the top of bottom of all pages.

While it can be configured with arbitrary links, by default it automatically
inserts a Table Of Contents entry, similar to the old python documentation format.
This is mainly useful when the title link in the navigation bar has been redirected
to page other than Table Of Contents (such as allowed by the ``roottarget``
option of the :doc:`/cloud_theme`).

Configuration
=============
By default, this inserts a "TOC" link to the right of any "Next" or "Previous"
buttons, which points to your sphinx master document.

It can be customized using the following options in ``conf.py``:

``relbar_links``

    Optional list of links to insert, defaults to ``["toc"]``.
    Each list item can have one of four formats:

    * A dictionary containing the keys ``"page"`` and ``"label"``,
      as well as optional ``"key"`` and ``"title"`` keywords.

      "page" should be the an absolute path to your page (e.g. "contents",
      or "lib/mypage"). "label" is the text that will be displayed on the link,
      "title" is the tooltip set for the link, and "key" is the accesskey
      for selecting the link.

    * A 2-element tuple of ``(page, label)``

    * A 4-element tuple of ``(page, title, key, label)``

    * The predefined string ``"toc"``, which is replaced with a table
      of contents link to your master document.

``relbar_links_after``

    When inserting the links, this extension will place them as far
    left as possible, so long as they are to the right of any links
    that are included in this list. Defaults to ``["next", "previous"]``.
"""
#=========================================================
# imports
#=========================================================
# core
from __future__ import absolute_import, division, print_function
import logging
log = logging.getLogger(__name__)
# site
from sphinx.locale import _
# pkg
from cloud_sptheme import __version__
from cloud_sptheme.utils import unicode
# local
__all__ = [
    "setup"
]

#=========================================================
# helpers
#=========================================================

# indices for looking up link-row elements in relbar entry
# TODO: find authoriative source for this info.
NAME = 0  # name of document
TITLE = 1  # used for [title] attribute
ACCESS_KEY = 2  # used for [accesskey] attribute
LABEL = 3  # used for link content (NOTE: has already been run through translation)


def parse_entry(app, entry):
    """
    normalize entry representation from config
    """
    if entry == "toc":
        master_doc = app.config.master_doc
        return master_doc, _("Table Of Contents"), "C", _("toc")
    if isinstance(entry, (list, tuple)):
        if len(entry) == 4:
            return entry
        if len(entry) == 2:
            page, label = entry
            return page, label, None, label
    elif isinstance(entry, dict):
        page = entry.get("page")
        label = entry.get("label")
        if not (page and label):
            raise ValueError("dict must define at least 'page' and "
                             "'label': %r" % (entry,))
        title = entry.get("title") or label
        key = entry.get("key")
        return page, title, key, label
    raise ValueError("expected dict or 2/4-element sequence: %r" % (entry,))

#=========================================================
# hooks
#=========================================================

def insert_relbar_links(app, pagename, templatename, ctx, event_arg):
    # check for relbar link list
    # NOTE: not defined for some pages of some builders,
    #       e.g. json builder's "genindex" page.
    if "rellinks" not in ctx:
        return
    links = ctx['rellinks']

    # get list of links to insert
    custom_links = [parse_entry(app, entry)
                    for entry in app.config.relbar_links]

    # remove existing links with same label
    # NOTE: unicode() wrapper is to resolve any _TranslationProxy() instances
    custom_labels = set(unicode(entry[LABEL]) for entry in custom_links)
    idx = 0
    while idx < len(links):
        if unicode(links[idx][LABEL]) in custom_labels:
            del links[idx]
        else:
            idx += 1

    # insert custom links to right of any "next" / "previous" links
    # NOTE: 'links' is in right to left order!
    after_labels = set(unicode(_(label)) for label in app.config.relbar_links_after)
    for pos, entry in enumerate(links):
        if unicode(entry[LABEL]) in after_labels:
            break
    else:
        pos = len(links)

    # insert links
    links[pos:pos] = custom_links

#=========================================================
# sphinx entrypoint
#=========================================================
def setup(app):
    """
    setup extension for sphinx run
    """

    # let user customize list of links
    app.add_config_value('relbar_links', ["toc"], 'env')
    app.add_config_value('relbar_links_after', ["next", "previous"], 'env')

    # attach rewrite-hook
    app.connect('html-page-context', insert_relbar_links)

    # identifies the version of our extension
    return {'version': __version__}

#=========================================================
# eof
#=========================================================



"""
cloud_sptheme.ext.autoattribute_search_bases -- monkeypatches autodoc so ``autoattribute`` searches base classes for attr doc.
"""
#=============================================================================
# imports
#=============================================================================
# core
import logging; log = logging.getLogger(__name__)
# site
# pkg
from cloud_sptheme import __version__
from cloud_sptheme.utils import patchapplier, monkeypatch
# local
__all__ = [
    "setup",
]

#=============================================================================
# patch
#=============================================================================
@patchapplier
def _patch_autoattribute():
    from sphinx.ext.autodoc import AttributeDocumenter, ModuleAnalyzer, PycodeError

    @monkeypatch(AttributeDocumenter)
    def add_content(_wrapped, self, *args, **kwds):
        if not self._datadescriptor and self.analyzer and self.objpath:
            attr_docs = self.analyzer.find_attr_docs()
            key = ('.'.join(self.objpath[:-1]), self.objpath[-1])
            if key not in attr_docs:
                # look for parent class w/ correct attr
                if hasattr(self.parent, "__mro__"):
                    for basecls in self.parent.__mro__[1:]:
                        try:
                            analyzer = ModuleAnalyzer.for_module(basecls.__module__)
                            base_attr_docs = analyzer.find_attr_docs()
                        except PycodeError as err:
                            continue
                        # FIXME: need qualname or equivalent for basecls
                        base_key = (basecls.__name__, self.objpath[-1])
                        if base_key in base_attr_docs:
                            # insert data into existing analyzer
                            # XXX: might be prettier way to handle this,
                            #      (esp so actual source file was reported)
                            #      but would have to duplicate much of add_content()
                            attr_docs[key] = base_attr_docs[base_key]
                            break
        return _wrapped(self, *args, **kwds)

#=============================================================================
# sphinx entry point
#=============================================================================
def setup(app):
    # don't apply our patch unless actually loaded by sphinx
    _patch_autoattribute()

    # identifies the version of our extension
    return {'version': __version__}

#=============================================================================
# eoc
#=============================================================================



"""
cloud_sptheme.ext.docfield_markup -- monkeypatches sphinx to allow ``~`` in docfields.
"""
#=============================================================================
# imports
#=============================================================================
# core
import logging; log = logging.getLogger(__name__)
# site
# pkg
from cloud_sptheme import __version__
from cloud_sptheme.utils import patchapplier, monkeypatch
# local
__all__ = [
    "setup",
]

#=============================================================================
# patch
#=============================================================================
@patchapplier
def _patch_docfield():
    from sphinx.util.docfields import Field, nodes, addnodes

    # NOTE: would like to just wrap make_xref(), but have to override first arg (rawsource) of
    #       pending_xref() call, so have to replicate the code.

    @monkeypatch(Field)
    def make_xref(_wrapped, self, rolename, domain, target, innernode=addnodes.literal_emphasis,
                  contnode=None, env=None):
        #------------------------------------------------------------------------
        # custom prep work before real sphinx code.
        #------------------------------------------------------------------------

        # given raw text in 'target', parse into 'title' and effective 'target'
        # NOTE: this tries to replicate the markup convention used in PyXRefRole.process_link()
        rawtext = title = target
        if rawtext.startswith("~"):
            # if the first character is a tilde, don't display the module/class
            # parts of the content
            target = target.lstrip("~")
            title = target.rpartition(".")[2]

        # create wrapper for innernode(title,title) calls, which honors custom title.
        # NOTE: doing as wrapper so code below will match sphinx as closely as possible.
        #       could instead just plug this in as contnode (if not already set)
        node_type = innernode

        def innernode(*ignored):
            if issubclass(node_type, nodes.Text):
                # Text classes want rawtext second
                return node_type(title, rawtext)
            else:
                # Element classes want rawtext first
                return node_type(rawtext, title)

        #------------------------------------------------------------------------
        # remainder of this should match sphinx code exactly, except where noted.
        #------------------------------------------------------------------------
        if not rolename:
            return contnode or innernode(target, target)
        # NOTE: passing <title> as first arg, sphinx passes ''
        refnode = addnodes.pending_xref(title, refdomain=domain, refexplicit=False,
                                        reftype=rolename, reftarget=target)
        refnode += contnode or innernode(target, target)
        if env:
            env.domains[domain].process_field_xref(refnode)
        return refnode

#=============================================================================
# sphinx entrypoint
#=============================================================================
def setup(app):
    # don't apply our patch unless actually loaded by sphinx
    _patch_docfield()

    # identifies the version of our extension
    return {'version': __version__}

#=============================================================================
# eoc
#=============================================================================



"""cloud_sptheme.ext.issue_tracker - added ":issue:" text role to link to issue tracker"""
#===========================================================================
# imports
#===========================================================================
#core
import re
#site
from docutils import nodes
from docutils.parsers.rst.roles import set_classes
#pkg
from cloud_sptheme import __version__
#local

#===========================================================================
# issue role
#===========================================================================

def make_error(inliner, rawtext, line, value):
    "generate error node and msg"
    msg = inliner.reporter.error(value, line=line)
    node = inliner.problematic(rawtext, rawtext, msg)
    return [node], [msg]

def get_issue_tracker_title(config):
    "retrieve issue_tracker_title template"
    return getattr(config, "issue_tracker_title", None) or "issue {issue}"

def get_issue_tracker_url(config):
    "retrieve issue_tracker_url template, replacing aliases"
    template = getattr(config, "issue_tracker_url", None)
    if not template:
        # causes :issue:`xx` to be replaced with label instead of url.
        return None

    elif template.startswith("bb:"):
        # parse "bb:<project>", and replace with bitbucket url
        project = template[3:].strip("/")
        return "https://bitbucket.org/" + project + "/issues/{issue}"

    elif template.startswith("gh:"):
        # parse "gh:<project>", and replace with github url
        project = template[3:].strip("/")
        return "https://github.com/" + project + "/issues/{issue}"

    else:
        # assume it contains {issue} and possibly {title}
        return template

# pattern allows inside :issue: text roles
issue_re = re.compile("""
    ^
    (?:
        (?P<title>[^<]+)
        \s*
        <
        (?P<issue1>\d+)
        >
    |
        (?P<issue2>\d+)
    )
    $
    """, re.X)

def issue_role(name, rawtext, text, line, inliner, options={}, content=[]):
    "generate link to an issue"
    #NOTE:
    #   name - role name in doc, should be 'issue'
    #   rawtext - text of entire node
    #   text - contents of role
    #   lineno
    #   inliner - ???
    #   options - ???
    #   content - ???
    # returns (nodes, messages)

    # extract title & issue number from text
    m = issue_re.match(text)
    if m:
        issue = int(m.group("issue1") or m.group("issue2"))
        title = m.group("title")
    else:
        return make_error(inliner, rawtext, line,
                          "Invalid issue identifier: %r" % (text,))

    # get url template from config, resolve aliases
    config = inliner.document.settings.env.app.config
    url_template = get_issue_tracker_url(config)
    title_template = get_issue_tracker_title(config)

    # generate replacement node
    if not title:
        title = title_template.format(issue=issue)
    set_classes(options)
    clist = options.setdefault('classes',[])
    clist.append("issue")
    if url_template:
        url = url_template.format(issue=issue, title=title)
        node = nodes.reference(rawtext, title, refuri=url, **options)
    else:
        node = nodes.emphasis(rawtext, title, **options)
    return [node], []

#===========================================================================
# init
#===========================================================================

def setup(app):
    "install plugin"

    app.add_config_value('issue_tracker_url', None, 'env')
    app.add_config_value('issue_tracker_title', None, 'env')
    app.add_role('issue', issue_role)

    # identifies the version of our extension
    return {'version': __version__}

#===========================================================================
# eof
#===========================================================================



"""cloud_sptheme.ext.escaped_samp_literals - allow escaping { and } in samp."""
#=============================================================================
# imports
#=============================================================================
# core
# site
from docutils import nodes, utils
# pkg
from cloud_sptheme import __version__
from cloud_sptheme.utils import u, ru
# local
__all__ = [
    "setup",
]
#=============================================================================
# replacement samp role parser
#=============================================================================
def emph_literal_role(typ, rawtext, text, lineno, inliner,
                      options={}, content=[]):
    """replacement for sphinx's ``:samp:`` role handler.
    this is a bit stricter in it's parsing, and allows escaping of literal
    ``{`` and ``}`` characters.
    """
    def make_error(pos, value):
        value = "%s at char %d of %s" % (value, pos, rawtext)
        msg = inliner.reporter.error(value, line=lineno)
        prb = inliner.problematic(rawtext, rawtext, msg)
        return [prb], [msg]
    text = utils.unescape(text)
    retnode = nodes.literal(role=typ.lower(), classes=[typ])
    buffer = u("") # contains text being accumulated for next node
    in_escape = False # True if next char is part of escape sequence
    in_var = False # True if parsing variable section instead of plain text
    var_start = None # marks start of var section if in_var is True
    i = 0
    for c in text:
        i += 1
        if in_escape:
            # parse escape sequence
            if c in ru("{}\\"):
                buffer += c
                in_escape = False
            else:
                return make_error(i-2, "unknown samp-escape '\\\\%s'" % (c,))
        elif c == ru("\\"):
            # begin escape sequence
            in_escape = True
            i += 1 # account for extra escape char in rawtext
        elif in_var:
            # parsing variable section
            if c == u("{"):
                return make_error(i, "unescaped '{'")
            elif c == u("}"):
                # finalize variable section, return to plaintext
                if not buffer:
                    return make_error(i-1, "empty variable section")
                retnode += nodes.emphasis(buffer, buffer)
                buffer = u("")
                in_var = False
            else:
                buffer += c
        else:
            # parsing plaintext section
            if c == u("{"):
                # finalize plaintext section, start variable section
                if buffer:
                    retnode += nodes.Text(buffer, buffer)
                buffer = u("")
                in_var = True
                var_start = i
            elif c == u("}"):
                return make_error(i, "unescaped '}'")
            else:
                buffer += c
    if in_escape:
        return make_error(i, "unterminated samp-escape sequence")
    elif in_var:
        return make_error(var_start, "unterminated variable section")
    elif buffer:
        retnode += nodes.Text(buffer, buffer)
    return [retnode], []

#=============================================================================
# sphinx entry point
#=============================================================================
def setup(app):
    # register our handler to overrride sphinx.roles.emph_literal_role
    from docutils.parsers.rst import roles
    import sphinx.roles as mod
    names = [
        key for key,value in mod.specific_docroles.items()
        if value is mod.emph_literal_role
    ]
    for name in names:
        roles.register_local_role(name, emph_literal_role)

    # identifies the version of our extension
    return {'version': __version__}

#=============================================================================
# eof
#=============================================================================



"""cloud_sptheme.ext.perpage -- override sphinx config per-page

* perpage_html_logo -- glob map (ala html_sidebars), used to change sidebar logo per-page

.. todo::
    document this extension
"""
#=============================================================================
# imports
#=============================================================================
# core
import os.path
import re
import logging; log = logging.getLogger(__name__)
# site
from sphinx.util.matching import patmatch
#pkg
from cloud_sptheme import __version__
# local
#=============================================================================
# helpers
#=============================================================================
def _rank_pattern(pattern):
    """return sorting key for prioritizing which glob pattern should match"""
    # TODO: add more ways to distinguish patterns if both have wildcards
    return not any(char in pattern for char in '*?[')

def bestmatch(patmap, source, default=None, param="source"):
    """return best match given a dictionary mapping glob pattersn -> values"""
    best = None
    best_rank = None
    for pattern in patmap:
        if not patmatch(pattern, source):
            continue
        cur_rank = _rank_pattern(pattern)
        if best is None or cur_rank < best_rank:
            best = pattern
            best_rank = cur_rank
        elif cur_rank == best_rank:
            raise KeyError("%s %r matches too many patterns: %r and %r" %
                           (param, source, best, pattern))
    if best is None:
        return default
    else:
        return patmap[best]

#=============================================================================
# sphinx hooks
#=============================================================================
def perpage_html_logo(app, pagename, templatename, ctx, event_arg):
    """helper to override sidebar logo per-page"""
    patmap = getattr(app.config, "perpage_html_logo", {})
    logo = bestmatch(patmap, pagename, ctx.get("logo"), param="pagename")
    if logo is None:
        ctx.pop("logo", None)
    else:
        ctx['logo'] = logo

# NOTE: this works, just don't have a use for it yet
##def perpage_html_theme_options(app, pagename, templatename, ctx, event_arg):
##    patmap = getattr(app.config, "perpage_html_theme_options")
##    if not patmap:
##        return
##    values = bestmatch(patmap, pagename, None, param="pagename")
##    if values:
##        for key, value in values.items():
##            # TODO: validate 'key' is valid theme option
##            ctx['theme_' + key] = value

#=============================================================================
# sphinx init
#=============================================================================
def setup(app):
    app.add_config_value('perpage_html_logo', None, 'env')
    ##app.add_config_value('perpage_html_theme_options', None, 'env')
    app.connect('html-page-context', perpage_html_logo)
    ##app.connect('html-page-context', perpage_html_theme_options)

    # identifies the version of our extension
    return {'version': __version__}

#=============================================================================
# eof
#=============================================================================



"""
==================================================================
:mod:`cloud_sptheme.ext.auto_redirect` -- Redirect Deprecated URLs
==================================================================

.. versionadded:: 1.9

Overview
========
This extension is helpful for when html documentation has been relocated
to a new host; e.g. moving from ``pythonhosted.org`` to ``readthedocs.io``.
Once enabled, it adds a helpful "Documentation has moved" message to the top
of every page, and automatically redirects the user as well.

Configuration
=============
This extension looks for the following config options:

    ``auto_redirect_subject``
    
        subject to insert into message.
        defaults to ``The {project name} documentation``.

    ``auto_redirect_domain_url``
        
        url to redirect user to.
        no message or redirect will happen if this isn't set.

    ``auto_redirect_domain_footer``
    
        optional footer text to append to message

Internals
=========
This should work with other sphinx themes, the js & css is (mostly) generic.

.. todo:: 

    the "redirect to exact page" part of JS code currently only works
    with this theme; would like to fix that.
    
.. todo::

    support configuring redirects to other pages within same documentation.

.. todo::

    if user has dismissed "auto redirect", remember via cookie?
"""
#=============================================================================
# imports
#=============================================================================
# core
import os
# site
from docutils.nodes import raw
from sphinx.builders.html import StandaloneHTMLBuilder
from sphinx.util.pycompat import htmlescape
# pkg
from cloud_sptheme import __version__, _root_dir, is_cloud_theme
from cloud_sptheme.utils import add_static_file
# local
__all__ = [
    "setup",
]

#=============================================================================
# patch builder to copy css file (if needed)
#=============================================================================

ext_static_dir = os.path.join(_root_dir, "ext", "static")

cloud_base_js_path = os.path.join(_root_dir, "themes", "cloud", "static", "cloud_base.js")

class RedirectHelper(object):

    template_path = os.path.join(ext_static_dir, "auto_redirect.html_t")
    css_path = os.path.join(ext_static_dir, "auto_redirect.css")

    def __init__(self, app):
        self.app = app

    @property
    def for_html(self):
        return isinstance(self.app.builder, StandaloneHTMLBuilder)

    @property
    def domain_subject(self):
        config = self.app.config
        return config.auto_redirect_domain_subject or \
               "The %s documentation" % (config.project,)
    @property
    def domain_url(self):
        return self.app.config.auto_redirect_domain_url

    @property
    def domain_root(self):
        root = self.app.config.auto_redirect_domain_root
        # XXX: use urljoin here?
        if not root:
            return self.domain_url
        if root[:1] == "/" and root[:2] != "//":
            return self.domain_url + root
        return root

    @property
    def domain_footer(self):
        return self.app.config.auto_redirect_domain_footer

    @property
    def timeout(self):
        return self.app.config.auto_redirect_timeout or 10


def install_static_assets(app):
    # skip unless html builder & redirect configured
    helper = RedirectHelper(app)
    if not helper.for_html or not helper.domain_url:
        return

    # make sure JS is added
    if not is_cloud_theme(app.config.html_theme):
        add_static_file(app.builder, cloud_base_js_path, javascript=True)

    # make sure needed css styling gets included when building html
    add_static_file(app.builder, helper.css_path, stylesheet=True)


# FIXME: This doesn't get called for auto-generated pages, e.g. search page & genindex.
#        Need to modify this code to get run as part of builder.handle_page() call,
#        such as via event that insert_canonical_link() is using below.
def prepend_redirect(app, doctree):

    """
    called after doctree has been read;
    inserted "Raw" node with our content.
    """
    # skip unless configured, and using html builder
    helper = RedirectHelper(app)
    if not helper.for_html or not helper.domain_url:
        return

    # render fragment into insert
    builder = app.builder
    renderer = builder.templates
    ctx = dict(app=app, redirector=helper)
    with open(helper.template_path) as fh:
        content = renderer.render_string(fh.read(), ctx)

    # insert raw node w/ content
    node = raw(content, content, format="html")
    doctree.insert(0, node)


def insert_canonical_link(app, pagename, templatename, context, doctree):
    # skip unless html builder & redirect configured
    helper = RedirectHelper(app)
    root_url = helper.domain_root
    if not helper.for_html or not root_url:
        return

    # XXX: any gotchas w/ generating url from pagename this way?
    url = "%s/%s.html" % (root_url.rstrip("/"), pagename)

    # insert "canonical" link to help search engines
    # XXX: skipping this if another ext sets this?
    # XXX: split this out under separate name / extension?
    meta = context.get("metatags", "")
    context['metatags'] = meta + '<link rel="canonical" href="%s" />\n' % htmlescape(url, True)

#=============================================================================
# register extension
#=============================================================================
def setup(app):
    #
    # register config vars
    #

    # title to use when referring to project.
    # defaults to "The {project} documentation"
    app.add_config_value("auto_redirect_domain_subject", None, "html")

    # base url to redirect to -- if not set, extension does nothing.
    app.add_config_value("auto_redirect_domain_url", None, "html")

    # optional url suffix to add before adding per-page redirects.
    app.add_config_value("auto_redirect_domain_root", None, "html")

    # optional html to append to message
    app.add_config_value("auto_redirect_domain_footer", None, "html")

    # seconds before redirecting via JS
    app.add_config_value("auto_redirect_timeout", 10, "html")

    #
    # connect events
    # XXX: can we connect events on the fly? could do single check when builder init'd,
    #      and attach events there as needed.  could then make everything a method
    #      of the redirect helper.
    #
    app.connect("builder-inited", install_static_assets)
    app.connect("doctree-read", prepend_redirect)
    app.connect("html-page-context", insert_canonical_link)

    #
    # identifies the version of our extension
    #
    return {'version': __version__}

#=============================================================================
# eof
#=============================================================================



"""
cloud_sptheme.tests -- test relbar_toc extension
"""
#=============================================================================
# imports
#=============================================================================
# core
from __future__ import absolute_import, division, print_function
import logging
log = logging.getLogger(__name__)
import os
# site
import mock
# pkg
from .utils import TestCase, unittest
# subject
from cloud_sptheme.ext import relbar_links
# local
__all__ = [
]

#=============================================================================
# extension test
#=============================================================================

# root_dir = os.path.abspath(os.path.join(__file__, *[".."]*3))
# docs_dir = os.path.join(root_dir, "docs")

# TODO: would like to test extensions *actual* behavior with sphinx,
#       and/or test our assumptions about format of ctx & ctx['rellinks'].
#       Have idea of running integration test by running sphinx,
#       and building either our docs, or some "test" docs;
#       and then examining the output.  not sure best way to approach
#       that though.
# from sphinx import main
# target_dir = self.makeTempDir()
# self.assertFalse(main(["sphinx", "-b", "html", "-E", docs_dir, target_dir]))

#=============================================================================
# utils tests
#=============================================================================
class UtilsTestCase(TestCase):

    def setUp(self):
        super(UtilsTestCase, self).setUp()

        # patch sphinx's _() to be noop
        patch = mock.patch.object(relbar_links, "_", side_effect=lambda s: s)
        self.addCleanup(patch.__exit__)
        patch.__enter__()

    def test_insert_no_links(self):
        """
        test insert hooks -- with no links present (e.g. json builder)
        """
        from cloud_sptheme.ext.relbar_links import insert_relbar_links
        app = mock.Mock()
        insert_relbar_links(app, None, None, {}, None)

    def test_insert_default(self):
        """
        test insert hooks -- with default links
        """
        from cloud_sptheme.ext.relbar_links import insert_relbar_links
        app = mock.Mock()
        app.config.master_doc = "test-contents"
        app.config.relbar_links = ["toc"]
        app.config.relbar_links_after = ["next", "previous"]
        links = [
            ("genindex", "General Index", "I", "genidx"),
            ("next-test-page", "Next Page", "N", "next"),
            ("previous-test-page", "Previous Page", "P", "previous"),
        ]
        insert_relbar_links(app, None, None, dict(rellinks=links), None)
        self.assertEqual(links, [
            ("genindex", "General Index", "I", "genidx"),
            ("test-contents", "Table Of Contents", "C", "toc"),
            ("next-test-page", "Next Page", "N", "next"),
            ("previous-test-page", "Previous Page", "P", "previous"),
        ])

    def test_insert_custom(self):
        """
        test insert hooks -- with custom links
        """
        from cloud_sptheme.ext.relbar_links import insert_relbar_links
        app = mock.Mock()
        app.config.relbar_links = [dict(page="my-page", label="my page")]
        app.config.relbar_links_after = ["previous"]
        links = [
            ("genindex", "General Index", "I", "genidx"),
            ("next-test-page", "Next Page", "N", "next"),
            ("previous-test-page", "Previous Page", "P", "previous"),
        ]
        insert_relbar_links(app, None, None, dict(rellinks=links), None)
        self.assertEqual(links, [
            ("genindex", "General Index", "I", "genidx"),
            ("next-test-page", "Next Page", "N", "next"),
            ("my-page", "my page", None, "my page"),
            ("previous-test-page", "Previous Page", "P", "previous"),
        ])

    def test_parse_entry(self):
        """
        test parse_entry() helper
        """
        from cloud_sptheme.ext.relbar_links import parse_entry
        app = mock.Mock()
        app.config.master_doc = "test-master-doc"

        #
        # dicts
        #
        self.assertRaises(ValueError, parse_entry, app, dict())

        self.assertRaises(ValueError, parse_entry, app, dict(page="page"))

        self.assertRaises(ValueError, parse_entry, app, dict(label="label"))

        self.assertEqual(parse_entry(app, dict(page="page", label="label")),
                         ("page", "label", None, "label"))

        self.assertEqual(parse_entry(app, dict(page="page", title="title", key="key", label="label")),
                         ("page", "title", "key", "label"))

        #
        # tuples
        #
        self.assertRaises(ValueError, parse_entry, app, ())

        self.assertRaises(ValueError, parse_entry, app, ("a",))

        self.assertEqual(parse_entry(app, ("page", "label")),
                         ("page", "label", None, "label"))

        self.assertRaises(ValueError, parse_entry, app, ("a", "b", "c"))

        self.assertEqual(parse_entry(app, ("page", "title", "key", "label")),
                         ("page", "title", "key", "label"))

        #
        # other
        #
        self.assertEqual(parse_entry(app, "toc"),
                         ("test-master-doc", "Table Of Contents", "C", "toc"))

#=============================================================================
# eof
#=============================================================================




"""
cloud_sptheme setup script
"""
#=============================================================================
# init script env -- ensure cwd = root of source dir
#=============================================================================
import os
root_dir = os.path.abspath(os.path.join(__file__, ".."))
os.chdir(root_dir)

#=============================================================================
# imports
#=============================================================================
import setuptools
import sys

#=============================================================================
# init setup options
#=============================================================================
opts = dict(
    #==================================================================
    # sources
    #==================================================================
    packages=setuptools.find_packages(root_dir),
    package_data={
	"cloud_sptheme": ["themes/*/*.*", "themes/*/static/*.*", "ext/static/*.*"]
    },
    zip_safe=False,

    #==================================================================
    # metadata
    #==================================================================
    name="cloud_sptheme",
    # NOTE: 'version' set below
    author="Eli Collins",
    author_email="elic@assurancetechnologies.com",
    license = "BSD",

    url = "https://bitbucket.org/ecollins/cloud_sptheme",
    download_url = "https://pypi.python.org/pypi/cloud_sptheme",

    install_requires=[
        "sphinx>=1.4"
    ],

    entry_points={
        'sphinx_themes': [
            'path = cloud_sptheme:get_theme_dir',
        ],
    },

    #==================================================================
    # details
    #==================================================================
    description= 
    "a nice sphinx theme named 'Cloud', and some related extensions",

    long_description="""\
This is a small package containing a Sphinx theme named "Cloud",
along with some related Sphinx extensions. To see an example
of the theme in action, check out it's documentation
at `<https://cloud-sptheme.readthedocs.io>`_.
    """,

    keywords="sphinx extension theme",

    classifiers="""
Development Status :: 5 - Production/Stable
Framework :: Sphinx :: Extension
Framework :: Sphinx :: Theme
Intended Audience :: Developers
License :: OSI Approved :: BSD License
Operating System :: OS Independent
Programming Language :: Python :: 2.6
Programming Language :: Python :: 2.7
Programming Language :: Python :: 3
Programming Language :: Python :: 3.3
Programming Language :: Python :: 3.4
Programming Language :: Python :: 3.5
Programming Language :: Python :: 3.6
Topic :: Documentation
Topic :: Software Development :: Documentation
""".strip().splitlines(),

    #==================================================================
    # custom setup
    #==================================================================
    script_args=sys.argv[1:],
    cmdclass={},
)

#=============================================================================
# set version string
#=============================================================================

# pull version string from package
from cloud_sptheme import __version__ as version

# append hg revision to builds
stamp_build = False
if stamp_build:
    from cloud_sptheme._setup.stamp import (
        as_bool, append_hg_revision, stamp_distutils_output,
        install_build_py_exclude, set_command_options
    )

    # add HG revision to end of version
    if as_bool(os.environ.get("SETUP_TAG_RELEASE", "yes")):
        version = append_hg_revision(version)

    # subclass build_py & sdist to rewrite source version string,
    # and clears stamp_build flag (above) so this doesn't run again.
    stamp_distutils_output(opts, version)

    # exclude '._setup' package from builds, only needed for sdist
    install_build_py_exclude(opts)
    set_command_options(opts, "build_py",
        exclude_packages=["cloud_sptheme._setup"],
    )

opts['version'] = version

#=============================================================================
# run setup
#=============================================================================
setuptools.setup(**opts)

#=============================================================================
# eof
#=============================================================================



"""cloud_sptheme.ext.index_styling - improves css styling for genindex"""
import logging; log = logging.getLogger(__name__)
import re
from jinja2 import Markup as literal, escape
from cloud_sptheme import __version__

prefix = r"^(?P<name>.*)\("
suffix = r"\)$"
_attr_re = re.compile(prefix + r"(?P<left>)(?P<loc>.*)(?P<right> attribute)" + suffix)
_meth_re = re.compile(prefix + r"(?P<left>)(?P<loc>.*)(?P<right> method)" + suffix)
_fc_re = re.compile(prefix + r"(?P<left>class in |in module )(?P<loc>.*)(?P<right>)" + suffix)
_mod_re = re.compile(prefix + r"module" + suffix)

def format_index_name(name):
    while True:
        m = _attr_re.match(name)
        if m:
            name, left, loc, right = m.group("name","left", "loc", "right")
            type = "attribute"
            break
        m = _meth_re.match(name)
        if m:
            name, left, loc, right = m.group("name","left", "loc", "right")
            type = "method"
            break
        m = _fc_re.match(name)
        if m:
            name, left, loc, right = m.group("name","left", "loc", "right")
            if left.startswith("class"):
                type = "class"
            else:
                type = "function"
            break
        m = _mod_re.match(name)
        if m:
            name = m.group("name")
            left = "module"
            loc = right = ''
            type = "module"
            break
        return name
    if loc:
        loc = literal('<span class="location">') + escape(loc) + literal("</span>")
    cat = left + loc + right
    return escape(name) + literal('<span class="category ' + type + '">') + escape(cat) + literal("</span>")

def mangle_index(app, pagename, templatename, ctx, event_arg):
    if pagename != "genindex":
        return
    fmt = format_index_name
    for key, entries in ctx['genindexentries']:
        for idx, entry in enumerate(entries):
            name, data = entry
            entries[idx] = fmt(name), data
            # NOTE: data is list of [links, subitems, some_key],
            #       though 'key' not added until sphinx 1.4 (1.3?)
            subitems = data[1]
            for idx, entry in enumerate(subitems):
                name, links = entry
                subitems[idx] = fmt(name), links

def setup(app):
    app.connect('html-page-context', mangle_index)

    # identifies the version of our extension
    return {'version': __version__}
